require 'crowdkit/api/arguments'
require 'crowdkit/api/factory'
require 'crowdkit/api/request_methods'
require 'crowdkit/api/response_wrapper'

module Crowdkit
  class API
    extend Forwardable
    include RequestMethods
    attr_accessor :stored_params
    attr_reader :client
    attr_writer :auto_pagination

    #We delegate last_response to the attr_accessor of the top level Client
    def_delegators :client, :config, :last_response, :last_response=

    #This should only be called by API Factory to ensure _client
    #gets set.  We override this in Client as it's the root of the API.
    def initialize(options, &block)
      @client = options.delete(:_client)
      @stored_params = options
      with(options)
    end

    def auto_pagination
      @auto_pagination || config.auto_paginate
    end

    # Acts as setter and getter for api requests arguments parsing.
    #
    # Returns Arguments instance.
    #
    def arguments(args=(not_set = true), options={}, &block)
      if not_set
        @arguments
      else
        @arguments = Arguments.new(self, options.symbolize_keys).parse(*args, &block)
      end
    end

    # Set an option to a given value
    #
    # @param [String] option
    # @param [Object] value
    # @param [Boolean] ignore_setter
    #
    # @return [self]
    #
    # @api public
    def set(option, value=(not_set=true), ignore_setter=false, &block)
      raise ArgumentError, 'value not set' if block and !not_set
      return self if !not_set and value.nil?

      if not_set
        set_options option
        return self
      end

      if respond_to?("#{option}=") and not ignore_setter
        return __send__("#{option}=", value)
      end

      define_accessors option, value
      self
    end

    #overridden in children API's that have their own id's
    def id_key
      :job_id
    end

    # Scope for passing request required arguments.
    #
    def with(args)
      case args
      when Hash
        #This happens when the job id is passed early in the chain
        if (custom = args.delete(:_args)) && custom.first
          stored_params[id_key] = custom.first
          set id_key, custom.first
        end
        set args
      when Fixnum, /^(\d+|([a-z])([a-z]|\d)*)$/
        stored_params[id_key] = args
        set id_key, args
      else
        ::Kernel.raise ArgumentError, 'This api does not support passed in arguments'
      end
    end

    # Defines a namespace
    #
    # @param [Array[Symbol]] names
    #   the name for the scope
    #
    # @return [self]
    #
    # @api public
    def self.namespace(*names)
      options = names.last.is_a?(Hash) ? names.pop : {}
      names   = names.map(&:to_sym)
      name    = names.pop
      return if public_method_defined?(name)

      class_name = extract_class_name(name, options)
      define_method(name) do |*args, &block|
        options = args.last.is_a?(Hash) ? args.pop : {}
        options[:_args] = args
        options[:_client] = @client
        API::Factory.new(class_name, stored_params.merge(options), &block)
      end
      self
    end

    # Extracts class name from options
    #
    # @return [String]
    #
    # @api private
    def self.extract_class_name(name, options)
      if !options[:class_name]
        converted  = options.fetch(:full_name, name).to_s
        converted  = converted.split('_').map(&:capitalize).join
        class_name = options.fetch(:root, false) ? '': "#{self.name}::"
        class_name += converted
        class_name
      else
        options[:class_name]
      end
    end

    private

    # Set multiple options
    #
    # @api private
    def set_options(options)
      unless options.respond_to?(:each)
        raise ArgumentError, 'cannot iterate over value'
      end
      options.each { |key, value| set(key, value) }
    end

    # Define setters and getters
    #
    # @api private
    def define_accessors(option, value)
      setter = proc { |val|  set option, val, true }
      getter = proc { value }

      define_singleton_method("#{option}=", setter) if setter
      define_singleton_method(option, getter) if getter
    end

    # Dynamically define a method for setting request option
    #
    # @api private
    def define_singleton_method(method_name, content=Proc.new)
      (class << self; self; end).class_eval do
        undef_method(method_name) if method_defined?(method_name)
        if String === content
          class_eval("def #{method_name}() #{content}; end")
        else
          define_method(method_name, &content)
        end
      end
    end
  end # API
end # Crowdkit
