module Crowdkit
  class API
    module RequestMethods
      # Make an HTTP request
      #
      # @param url [String] The path, relative to {#api_endpoint}
      # @param options [Hash] Query and header params for request
      # @return [Sawyer::Resource]
      [:get,:post,:put,:patch,:delete,:head].each do |method|
        define_method("do_#{method}") do |url, data = {}, options = {}|
          request method, url, data, options
        end
      end

      def agent
        raise Crowdkit::Unauthorized unless config.access_token
        @agent ||= Sawyer::Agent.new(config.api_endpoint, sawyer_options)
      end

      private

      def request(method, path, data, options = {})
        if data.is_a?(Hash)
          options[:headers] = data.delete(:headers) || {}
          options[:headers][:accept] = accept if accept = data.delete(:accept)
          options[:query]   = data.delete(:query) || {}
          if [:get, :head, :delete].include?(method)
            data.each do |key, _|
              options[:query][key] = data.delete(key)
            end
          else
            options[:headers]['Content-Type'] = 'application/json'
          end
        end

        begin
          Crowdkit.last_response = self.last_response = agent.call(method, URI::Parser.new.escape(path.to_s), data, options)
          @retried = false
          ResponseWrapper.new(self.last_response, self.client)
        rescue Crowdkit::ServiceError => e
          Crowdkit.last_response = self.last_response = Sawyer::Response.new(agent, e.response)

          # The per second limit was exceeded, let's try again in a second
          if e.response.status == 429 && !@retried
            sleep(1)
            @retried = true
            retry
          end
          raise e
        end
      end

      # Executes the request, checking if it was successful
      #
      # @return [Boolean] True on success, false otherwise
      def boolean_from_response(method, path, options = {})
        response = request(method, path, options)
        response.status == 204
      rescue Crowdkit::NotFound
        false
      end

      def sawyer_options
        conn_opts = config.connection_options
        conn_opts[:builder] = config.stack
        { faraday: Faraday.new(conn_opts) }
      end
    end
  end
end
