module Crowdkit
  class API
    class ResponseWrapper
      attr_reader :sawyer, :client
      extend Forwardable
      include RequestMethods

      def_delegators :data, :empty?, :size, :include?, :length, :first, :last, :flatten, :include?, :keys, :to_s, :to_hash, :rels
      def_delegators :client, :last_response, :last_response=, :config, :auto_pagination
      def_delegators :each, :to_a, :to_ary, :to_enum

      def initialize(sawyer, client)
        @sawyer = sawyer
        @client = client
      end

      def data
        @sawyer.data
      end

      def headers
        @sawyer.headers
      end

      # Lookup an attribute from the body if hash, otherwise behave like array index.
      # Convert any key to string before calling.
      #
      def [](key)
        if self.data.is_a?(Array)
          self.data[key]
        else
          self.data.send(:"#{key}")
        end
      end

      def current_status
        if self.data["current_status"]
          do_get(self.data["current_status"].rels["self"].href)
        else
          nil
        end
      end

      # Iterate over each resource inside the body
      #
      def each(to_enum = true, &block)
        body_parts = self.data.respond_to?(:each) ? self.data : [self.data]
        if block_given?
          body_parts.each &block
          if self.auto_pagination && self.has_next_page?
            next_page.each &block
          end
        else
          res = self
          if self.auto_pagination && res.has_next_page?
            res = next_page
            body_parts.concat(res.each(false))
          end
          result = to_enum ? body_parts.to_enum : body_parts
        end
        #Reset auto pagination option
        client.auto_pagination = nil
        result
      end

      [:next, :last, :first, :prev].each do |rel|
        define_method("#{rel}_page") do
          @sawyer.rels[rel] && do_get(@sawyer.rels[rel].href)
        end
      end

      def has_next_page?
        @sawyer.rels[:next]
      end

      # Overwrite methods to hash keys
      #
      ['id', 'type'].each do |method_name|
        define_method(method_name) do
          self.[](method_name)
        end
      end

      def rate_limit_remaining
        self.headers["X-RateLimit-Remaining"]
      end

      # Check if body has an attribute
      #
      def has_key?(key)
        res = self.data.is_a?(Array) ? self.data.first : self.data
        res.key?(key.to_sym)
      end

      # Coerce any method calls for body attributes
      #
      def method_missing(method_name, *args, &block)
        if self.has_key?(method_name)
          self.[](method_name, &block)
        else
          super
        end
      end

      # Check if method is defined on the body
      #
      def respond_to?(method_name)
        if self.has_key?(method_name)
          true
        else
          super
        end
      end

      # Print only response body
      #
      def inspect
        self.data.inspect
      end
    end
  end
end
