module Crowdkit
  class Error < StandardError; end

  # Custom error class for errors made with the client itself
  class UserError < Error
    def generate_message(attributes)
      "\nProblem:\n #{attributes[:problem]}"+
      "\nSummary:\n #{attributes[:summary]}"+
      "\nResolution:\n #{attributes[:resolution]}"
    end
  end

  class RequiredParams < UserError
    def initialize(provided, required)
      super(
        generate_message(
          :problem => "Missing required parameters: #{provided.keys.join(', ')} provided for this request.",
          :summary => "Crowdkit checks the request parameters passed to ensure the CrowdFlower api is not hit unnecessairly.",
          :resolution => "Required parameters are: #{required.join(', ')}, make sure these are the ones you are using"
        )
      )
    end
  end

  class ValidationError < UserError
    def initialize(errors)
      super(
        generate_message(
          :problem => "Attempted to send request with nil arguments for #{errors.keys.join(', ')}.",
          :summary => 'Each request expects certain number of required arguments.',
          :resolution => 'Double check that the provided arguments are set to some value that is neither nil nor empty string.'
        )
      )
    end
  end

  class StatusError < UserError
    def initialize(status, timed_out_after = nil)
      if timed_out_after
        reason = "timed out"
        message = "it did not complete after #{timed_out_after} seconds"
      else
        reason = status.state == "killed" ? "was killed" : "failed"
        message = status.message
      end
      super(
        generate_message(
          :problem => "There was a problem while processing your request in the background.",
          :summary => "The process #{reason} because of: #{message}.",
          :resolution => "Try performing the operation again."
        )
      )
    end
  end

  # Custom error class for rescuing from all CrowdFlower errors
  class ServiceError < Error
    attr_reader :response

    # Returns the appropriate Crowdkit::Error sublcass based
    # on status and response message
    #
    # @param [Hash] response HTTP response
    # @return [Crowdkit::Error]
    def self.from_response(response)
      status  = response[:status].to_i
      body    = response[:body].to_s
      headers = response[:response_headers]

      if klass =  case status
                  when 400      then Crowdkit::BadRequest
                  when 401      then Crowdkit::Unauthorized
                  when 403      then Crowdkit::Forbidden
                  when 404      then Crowdkit::NotFound
                  when 406      then Crowdkit::NotAcceptable
                  when 409      then Crowdkit::Conflict
                  when 415      then Crowdkit::UnsupportedMediaType
                  when 422      then Crowdkit::UnprocessableEntity
                  when 429      then Crowdkit::TooManyRequests
                  when 400..499 then Crowdkit::ClientError
                  when 500      then Crowdkit::InternalServerError
                  when 501      then Crowdkit::NotImplemented
                  when 502      then Crowdkit::BadGateway
                  when 503      then Crowdkit::ServiceUnavailable
                  when 500..599 then Crowdkit::ServerError
                  end
        klass.new(response)
      end
    end

    def initialize(response=nil)
      @response = Faraday::Response.new(response)
      super(build_error_message)
    end

    # Documentation URL returned by the API for some errors
    #
    # @return [String]
    def documentation_url
      data[:_links][:documentation] if data.is_a?(Hash) && data[:_links]
    end

    # Array of validation errors
    # @return [Array<Hash>] Error info
    def errors
      if data && data.is_a?(Hash)
        data[:errors] || []
      else
        []
      end
    end

    private

    def data
      @data ||=
        if (body = @response.body) && !body.empty?
          if body.is_a?(String) &&
            @response.headers &&
            @response.headers[:content_type] =~ /json/

            Sawyer::Agent.serializer.decode(body)
          else
            body
          end
        else
          nil
        end
    end

    def response_error_summary
      return nil if errors.empty?

      summary = "\nError summary:\n\t"
      summary << data[:errors].join("\n\t")
      summary << "\n"
    end

    def build_error_message
      return nil if @response.env.nil?

      message =  "#{@response.env[:method].to_s.upcase} "
      message << "#{@response.env[:url]} - "
      message << "(#{@response.status}):"
      message << "#{response_error_summary}" unless response_error_summary.nil?
      message << " // See: #{documentation_url}" unless documentation_url.nil?
      message
    end
  end

  # Raised on errors in the 400-499 range
  class ClientError < ServiceError; end

  # Raised when CrowdFlower returns a 400 HTTP status code
  class BadRequest < ClientError; end

  # Raised when CrowdFlower returns a 401 HTTP status code
  class Unauthorized < ClientError; end

  # Raised when CrowdFlower returns a 403 HTTP status code
  class Forbidden < ClientError; end

  # Raised when CrowdFlower returns a 404 HTTP status code
  class NotFound < ClientError; end

  # Raised when CrowdFlower returns a 406 HTTP status code
  class NotAcceptable < ClientError; end

  # Raised when CrowdFlower returns a 409 HTTP status code
  class Conflict < ClientError; end

  # Raised when CrowdFlower returns a 414 HTTP status code
  class UnsupportedMediaType < ClientError; end

  # Raised when CrowdFlower returns a 422 HTTP status code
  class UnprocessableEntity < ClientError; end

  # Raised when CrowdFlower returns a 429 HTTP status code
  class TooManyRequests < ClientError; end

  # Raised on errors in the 500-599 range
  class ServerError < ServiceError; end

  # Raised when CrowdFlower returns a 500 HTTP status code
  class InternalServerError < ServerError; end

  # Raised when CrowdFlower returns a 501 HTTP status code
  class NotImplemented < ServerError; end

  # Raised when CrowdFlower returns a 502 HTTP status code
  class BadGateway < ServerError; end

  # Raised when CrowdFlower returns a 503 HTTP status code
  class ServiceUnavailable < ServerError; end

  # Raised when client fails to provide valid Content-Type
  class MissingContentType < ArgumentError; end
end
